local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "rr_player0"
local policeNames = {"rr_police0", "rr_police1"}
local suspectName = "rr_suspect0"
local waypointNames = {"waypoint0", "waypoint1", "waypoint2", "waypoint3", "waypoint4", "waypoint5"}
local finishName = "waypoint5"
local messages = {fail = "The suspect made it to the safe house.", win = "Good job! This fool's next stop is jail.", quickWin = "Good job! That was impressively fast.", skillWin = "Good job! Looks like you didn't need any help from roadblocks!"}

local speedLimit = 24 --Speed limit of lawful drivers
local zonesTotal = 20 --Number of spawn zones in the prefab
local suspectAggro = 1.2 --The aggression level of the suspect
local nearDist = 50 --The distance that you can arrest at
local countMax = 5 --Arrest countdown timer start value
local mainTimer = 0
local zoneCheck = 0
local hornHold = 0
local delay = 0
local funMode = 0
local activator = nil
local allowRoadblock = false
local setRoadblock = false
local stockRoadblock = 0
local escape = nil
local slowMotion = false

local Police = {}
local player = {}
local police = {}
local suspect = {}
local zones = {}

function Police:new(obj, damage, state, zone, timer)
	local self = setmetatable({}, Police)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.damage = damage or nil --Vehicle damage
	self.state = state or nil --Behavior states; none, cruise, ready, wait, block
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.timer = timer or nil --Timer, used for delays
	return self
end

local function arrestSuspect(vehicle) --Arrests the suspect with a countdown
	local countdown = 0
	vehicle.timer = vehicle.timer + 0.25
	if vehicle.timer % 1 == 0 then
		local countdown = countMax - math.modf(vehicle.timer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	if vehicle.timer > countMax then
		if mainTimer <= 60 then
			scenario.finish({msg = messages.quickWin})
		elseif stockRoadblock == 3 then
			scenario.finish({msg = messages.skillWin})
		else
			scenario.finish({msg = messages.win})
		end
	end
end

local function preSpawn(vehicle) --Spawns the vehicle under the spawnzone, preparing it for position and rotation
	local zoneIn = scenetree.findObject(tostring("spawnzone"..(player.zone + 2 - funMode)))
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	vehicle.obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z - 5, 0, 0, zoneRot.z, zoneRot.w)
	log("Z", "gull", vehicle.obj.name.." teleported to "..zoneIn.name)
	vehicle.state = "wait"
	vehicle.zone = zoneIn.name
	vehicle.timer = 0
end

local function postSpawn(vehicle) --Spawns the vehicle into the roadblock formation
	local ra = 0
	local zonePos = scenetree.findObject(vehicle.zone):getPosition()
	for i, v in ipairs(police) do
		if vehicle.obj.name == police[i].obj.name then
			ra = (i * 2) - 3
		end
	end
	local posOffset = vec3(vehicle.obj:getPosition()) - vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, 0.7071 * ra, 0.7071)) * 3.333
	local rotOffset = quatFromDir(vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, 0.7071 * ra, 0.7071)), vec3(0,0,1))
	vehicle.obj:setPositionRotation(posOffset.x, posOffset.y, zonePos.z + 0.5, 0, 0, rotOffset.z, rotOffset.w)
	if vehicle.damage >= 1000 then
			vehicle.obj:queueLuaCommand("vehicleResetted(false)")
		end
	vehicle.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
	vehicle.state = "block"
	vehicle.timer = 0
end

local function onScenarioUIReady()
	player = {obj = scenetree.findObject(playerName), pos = nil, vel = nil, dist = nil, markerDist = nil, zone = 0}
	suspect = {obj = scenetree.findObject(suspectName), pos = nil, vel = nil, timer = 0}
	for i, v in ipairs(policeNames) do
		police[i] = Police:new(scenetree.findObject(policeNames[i]), 0, "cruise", nil, 0)
	end
	for  i = 1, zonesTotal, 1 do
		zones[i] = "spawnzone"..(i - 1)
	end
	activator = scenetree.findObject("activator")
	activator.hidden = false
	allowRoadblock = false
	setRoadblock = false
	stockRoadblock = 3
	mainTimer = 0
	zoneCheck = 0
	hornHold = 0
	funMode = 0
	escape = nil
	slowMotion = false
end

local function onRaceStart()
	local arg = {vehicleName = suspect.obj.name, waypoints = waypointNames, aggression = suspectAggro}
	helper.setAiPath(arg)
	for i, v in ipairs(police) do
		local arg = {vehicleName = police[i].obj.name, waypoints = waypointNames, driveInLane = "on", routeSpeed = speedLimit, routeSpeedMode = "limit"}
		helper.setAiPath(arg)
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	suspect.pos = suspect.obj:getPosition()
	suspect.vel = map.objects[map.objectNames[suspect.obj.name]].vel:length()
	player.dist = (player.pos - suspect.pos):len()
	player.markerDist = (player.pos - activator:getPosition()):len()
	player.obj:queueLuaCommand("obj:queueGameEngineLua('zHorn = ('..electrics.values.horn..')')") --Creates a global variable for the horn state; 0 = off and 1 = on
	if zHorn == 1 then
		hornHold = hornHold + raceTickTime
		if mainTimer == 1 then
			helper.flashUiMessage("Roadblocks not ready yet!", 2, false)
		end
		if mainTimer <= 10 and hornHold == 5 then
			helper.flashUiMessage("Fun mode enabled!", 2, false)
			funMode = 1
		end
	end
	if player.markerDist <= 12 and allowRoadblock == false then
		helper.flashUiMessage("Roadblock units ready!", 3, false)
		for i, v in ipairs(police) do
			police[i].obj:queueLuaCommand("ai.setMode('stopping')")
		end
		activator.hidden = true
		allowRoadblock = true
	end
	if zHorn == 1 and allowRoadblock == true and setRoadblock == false then
		if stockRoadblock <= 0 then
			if allowRoadblock == true then
				helper.flashUiMessage("You used up all of your roadblock calls!", 2, false)
				allowRoadblock = false
			end
		end
		if stockRoadblock > 0 then
			helper.flashUiMessage("Roadblock deployed!", 2, false)
			setRoadblock = true
			if funMode == 0 then
				stockRoadblock = stockRoadblock - 1
			end
			player.zone = zoneCheck
		end
	end
	if player.dist <= nearDist and suspect.vel <= 3 then
		arrestSuspect(suspect)
	else
		suspect.timer = 0
	end
	for i, v in ipairs(police) do
		police[i].damage = map.objects[map.objectNames[police[i].obj.name]].damage
		if mainTimer * 2 == i then
			police[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)") --Delays the AI police sirens by a bit; sounds nicer!
		end
		if police[i].state == "wait" then
			if police[i].timer == 0.5 then
				postSpawn(police[i])
			end
			police[i].timer = police[i].timer + raceTickTime
		end
		if setRoadblock == true and police[i].state == "cruise" then
			police[i].state = "ready"
			police[i].timer = 0
		end
		if delay > 0 and police[i].state == "ready" then
			if police[i].timer >= 0.5 and police[i].obj.name == police[delay].obj.name then
				preSpawn(police[i])
				delay = -4
			end
			police[i].timer = police[i].timer + raceTickTime
		end
	end
	delay = delay + 1
	if delay > #police then
		delay = 1
	end
	if commands.isFreeCamera(0) then
		if slowMotion == false then
			bullettime.set(0.125)
			slowMotion = true
		end
	else
		if slowMotion == true then
			bullettime.set(1)
			slowMotion = false
		end
	end
end

local function onRaceResult()
	if escape == true then
		scenario.finish({failed = messages.fail})
	end
end

local function onRaceWaypointReached(data)
	if data.waypointName == finishName then
		if data.vehicleName == suspect.obj.name and escape == nil then
			escape = true
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then
			for i, v in ipairs(zones) do
				if data.triggerName == zones[i] then
					zoneCheck = i
				end
			end
			if zoneCheck > zonesTotal - 3 then
				zoneCheck = 0
			end
		end
	end
	if data.event == "exit" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then
			if setRoadblock == true and data.triggerName == police[1].zone then
				for i, v in ipairs(police) do
					police[i].state = "cruise"
				end
				setRoadblock = false
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onRaceWaypointReached = onRaceWaypointReached
M.onBeamNGTrigger = onBeamNGTrigger

return M